function (::BoozerFromFlux)(x::FluxCoordinates,
                            vmec::VmecSurface{T},
                            wmn::AbstractArray{VmecFourierData{T}},
                            boozerG::T,
                            boozerI::T;
                           ) where {T <: AbstractFloat}
  w = inverseTransform(x,wmn)
  λ = inverseTransform(x,vmec.λmn)
  jacFac = 1.0/(boozerG + vmec.iota[1]*boozerI)
  p = (w + boozerI*λ)*jacFac
  ψ = x.ψ
  χ = x.θ + λ + vmec.iota[1]*p
  ϕ = x.ζ + p
  return BoozerCoordinates(ψ,χ,ϕ)
end

function jacobian(::BoozerFromFlux,
                  v::FluxCoordinates,
                  vmec::VmecSurface{T},
                  wmn::AbstractVector{VmecFourierData{T}},
                  boozerG::T,
                  boozerI::T;
                 ) where {T <: AbstractFloat}
  jacFac = 1.0/(boozerG + vmec.iota[1]*boozerI)
  dλdθ = inverseTransform(v,vmec.λmn;deriv=:dθ)
  dλdζ = inverseTransform(v,vmec.λmn;deriv=:dζ)
  dpdθ = (inverseTransform(v,wmn;deriv=:dθ) - boozerI*dλdθ)*jacFac
  dpdζ = (inverseTransform(v,wmn;deriv=:dζ) - boozerI*dλdζ)*jacFac
  return (1+dλdθ)*(1+dpdζ) + (vmec.iota[1] - dλdζ)*dpdθ
end

function boozerWFunction(vmec::VmecSurface{T};
                        ) where {T <: AbstractFloat}
  #wmn = similar(vmec.bsubumn)
  boozerG = vmec.bsubvmn[1].cos
  boozerI = vmec.bsubumn[1].cos
  zeros_T = zeros(T,vmec.mnmax_nyq)
  wmn_cos = Array{T}(undef,vmec.mnmax_nyq)
  wmn_sin = similar(wmn_cos)
  for mn in eachindex(wmn_cos,wmn_sin)
    m = vmec.xm_nyq[mn]
    n = vmec.xn_nyq[mn]
    if m != 0 && n != 0
      wmn_cos[mn] = 0.5*(vmec.bsubumn[mn].cos/m - vmec.bsubvmn[mn].cos/n)
      wmn_sin[mn] = 0.5*(vmec.bsubumn[mn].sin/m - vmec.bsubvmn[mn].sin/n)
      #wmn[mn] = VmecFourierData{T}(m,n,0.5*(vmec.bsubumn[mn].cos/m - vmec.bsubvmn[mn].cos/n),zeroT,0.5*(vmec.bsubumn[mn].sin/m - vmec.bsubvmn[mn].sin/n),zeroT)
   elseif m != 0
      wmn_cos[mn] = vmec.bsubumn[mn].cos/m
      wmn_sin[mn] = vmec.bsubumn[mn].sin/m
      #wmn[mn] = VmecFourierData{T}(m,n,vmec.bsubumn[mn].cos/m,zeroT,vmec.bsubumn[mn].sin/m,zeroT)
    elseif n != 0
      wmn_cos[mn] = -vmec.bsubvmn[mn].cos/n
      wmn_sin[mn] = -vmec.bsubvmn[mn].sin/n
      #wmn[mn] = VmecFourierData{T}(m,n,-vmec.bsubvmn[mn].cos/n,zeroT,-vmec.bsubvmn[mn].sin/n,zeroT)
    else
      wmn_cos[mn] = zero(T)
      wmn_sin[mn] = zero(T)
      #wmn[mn] = VmecFourierData{T}(zeroT,zeroT,zeroT,zeroT,zeroT,zeroT)
    end
  end
    #=
    if vmec.xm_nyq[mn] != 0
      wmn[mn] = !vmec.lasym ?
      SVector(vmec.bsubumn[mn][1],vmec.bsubumn[mn][2],vmec.bsubumn[mn][3]/vmec.bsubumn[mn][1],0.0) :
      SVector(vmec.bsubumn[mn][1],vmec.bsubumn[mn][2],vmec.bsubumn[mn][3]/vmec.bsubumn[mn][1],0.0,vmec.bsubumn[mn][5]/vmec.subumn[mn][1],0.0)
      wmn[mn] = VmecFourierData{T}(vmec.bsubumn[mn].m,vmec.bsubumn[mn]
    elseif vmec.xn_nyq[mn] != 0
      wmn[mn] = !vmec.lasym ?
      SVector(vmec.bsubvmn[mn][1],vmec.bsubvmn[mn][2],-vmec.bsubvmn[mn][3]/vmec.bsubvmn[mn][2],0.0) :
      SVector(vmec.bsubvmn[mn][1],vmec.bsubvmn[mn][2],-vmec.bsubvmn[mn][3]/vmec.bsubvmn[mn][2],0.0,-vmec.bsubvmn[mn][5]/vmec.subvmn[mn][2],0.0)
      wmn[mn] = VmecFourierData{T}(vmec.xm_nyq[mn],vmec.xn_nyq[mn],-vmec.bsubvmn[mn].cos/vmec.xn_nyq[mn],zero(T),-vmec.subvmn[mn].sin/vmec.xn_nyq
    else
      wmn[mn] = VmecFourierData{T}(zero(T),zero(T),zero(T),zero(T),zero(T),zero(T))
    end
  end
  =#
  return StructArray{VmecFourierData{T}}((vmec.xm_nyq,vmec.xn_nyq,wmn_cos,zeros_T,wmn_sin,zeros_T)), boozerG, boozerI
end

function BoozerFourierData(len::Int,asym::Bool=false)
  if !asym
    return Vector{SVector{3,Float64}}(undef,len)
  else
    return Vector{SVector{4,Float64}}(undef,len)
  end
end

function boozerGrid(vmec::VmecSurface,
                    mBoozer::Int,
                    nBoozer::Int,
                    wmn::AbstractVector{VmecFourierData{T}},
                    boozerG::T,
                    boozerI::T;
                   ) where {T <: AbstractFloat}
  θ_range = range(0.0,step=2π/(2*(2*mBoozer+1)),length=2*(2*mBoozer+1))
  ζ_range = range(0.0,step=2π/(vmec.nfp*2*(2*nBoozer+1)),length=2(2*nBoozer+1))
  u = MagneticCoordinateGrid(FluxCoordinates,vmec.s*vmec.phi[end]/2π*vmec.signgs,θ_range,ζ_range)
  b = StructArray{BoozerCoordinates{T,T}}(undef,size(u))
  jac = Array{T,ndims(u)}(undef,size(u))
  @batch minbatch=16 for i in eachindex(u,b)
    b[i] = BoozerFromFlux()(u[i],vmec,wmn,boozerG,boozerI)
  end
  @batch minbatch=16 for i in eachindex(u,jac)
    jac[i] = jacobian(BoozerFromFlux(),u[i],vmec,wmn,boozerG,boozerI)
  end
  return u, b, jac
end

function boozerModes(mBoozer::Int,
                     nBoozer::Int,
                     nfp::Int;
                    )
  mnBoozer = nBoozer + 1 + (mBoozer-1)*(1 + 2*nBoozer)
  modes = Vector{SVector{2,Int}}(undef,mnBoozer)
  i = 1
  for n = 0:nfp:nfp*nBoozer
    modes[i] = SVector{2,Int}(0,n)
    i += 1
  end
  for m = 1:mBoozer-1
    for n = -nfp*nBoozer:nfp:nfp*nBoozer
      modes[i] = SVector{2,Int}(m,n)
      i += 1
    end
  end
  return mnBoozer, modes
end

function boozerCosineSpectrum(field::Symbol,
                              vmec::VmecSurface{T},
                              mBoozer::Int,
                              nBoozer::Int;
                             ) where {T}
  mnBoozer, modes = boozerModes(mBoozer,nBoozer,vmec.nfp)
  spectrum = BoozerFourierData(mnBoozer)
  scaleFactor = 2.0/(2*(2*mBoozer+1)*2*(2*nBoozer+1))
  wmn, boozerG, boozerI = boozerWFunction(vmec)
  #Set up the coordinates, field data and Jacobian of the transformation
  u, b, jac = boozerGrid(vmec,mBoozer,nBoozer,wmn,boozerG,boozerI)
  data = Matrix{Float64}(undef,size(u))
  for i = 1:length(u)
    data[i] = inverseTransform(u[i],getfield(vmec,field)) * jac[i]
  end
  @batch minbatch=128 for i = 1:length(modes)
    spectrum[i] = SVector{3,T}(modes[i][1],modes[i][2],
                          cosineTransform(modes[i][1],modes[i][2],b,data)*scaleFactor)
  end
  spectrum[1] = SVector(0,0,spectrum[1][3]*0.5)
  return spectrum
end

function boozerCosineSpectrum(field::Symbol,
                              surfaces::AbstractVector,vmec::Vmec{T},mBoozer::Int,nBoozer::Int) where {T <: AbstractFloat}
  spectrumData = Vector{Vector{SVector{3,Float64}}}(undef,length(surfaces))
  for i = 1:length(surfaces)
    vmecSurface = VmecSurface(surfaces[i],vmec)
    spectrumData[i] = boozerCosineSpectrum(field,vmecSurface,mBoozer,nBoozer)
  end
  return spectrumData
end

function component(spectrum::Vector{SVector{N,T}},m::Int,n::Int) where {N,T}
  res = BitArray(undef,length(spectrum))
  map!(i->(i[1]==m)*(i[2]==n),res,spectrum)
  index = findfirst(res.==1)
  return spectrum[index][3]
end

function quasisymmetry(mQS::Int,
                       nQS::Int,
                       mBoozer::Int,
                       nBoozer::Int,
                       vmec::VmecSurface{T};
                      ) where {T <: AbstractFloat}
  modB = boozerCosineSpectrum(:bmn,vmec,mBoozer,nBoozer)
  polModes = map(i->i[1],modB)
  torModes = map(i->i[2],modB)
  denom = abs(modB[1][3])
  asymPower = sum(map(i->i[3]^2,modB[2:end]))
  m = mQS
  n = nQS
  while m <= (mBoozer-1) && n <= vmec.nfp*nBoozer
    index = findfirst((polModes.==m).*(torModes.==n).==1)
    asymPower -= modB[index][3]^2
    m += mQS
    n += nQS
  end
  return asymPower > zero(typeof(asymPower)) ? sqrt(asymPower)/denom : zero(typeof(asymPower))
end

function quasisymmetry(mQS::Int,
                       nQS::Int,
                       mBoozer::Int,
                       nBoozer::Int,
                       vmecSurfaces::AbstractVector{VmecSurface{T}};
                      ) where {T}
  qsValues = Vector{T}(undef,length(vmecSurfaces))
  for i in eachindex(qsValues,vmecSurfaces)
    qsValues[i] = quasisymmetry(mQS,nQS,mBoozer,nBoozer,vmecSurfaces[i])
  end
  return qsValues
end

module QuasiSymmetry
  using VMEC
  using MPI

  """
      target_quasisymmetry(mQS::Int,nQS::Int,mBoozer::Int,nBoozer::Int,vmec::VmecSurface{T}) where {T <: AbstractFloat}

  Compute the square root of the sum of squares of the non-quasisymmetric modes
  normalized by the (0,0) mode.
  """
  function target_quasisymmetry(mQS::Int,nQS::Int,mBoozer::Int,nBoozer::Int,vmec::VmecSurface{T}) where {T <: AbstractFloat}
    return VMEC.quasisymmetry(mQS,nQS,mBoozer,nBoozer,vmec)
  end

  function target_quasisymmetry(mQS::Int,nQS::Int,mBoozer::Int,nBoozer::Int,surfaceLabels::AbstractVector{VmecSurface{T}}) where {T}
    targetValues = Vector{Float64}(undef,length(surfaceLabels))
    Threads.@threads for i = 1:length(surfaceLabels)
      targetValues[i] = target_quasisymmetry(mQS,nQS,mBoozer,nBoozer,surfaceLabels[i])
    end
    return sum(targetValues)
  end

  function computeTarget(args::Dict{Symbol,Any},vmecSurface::VmecSurface{T}) where {T}
    if args[:surfaceLabel] != vmecSurface.s
      throw(error("Input surface label and VmecSurface label do not match!"))
    end
    return target_quasisymmetry(args[:mQS],args[:nQS],args[:mBoozer],args[:nBoozer],vmecSurface)
  end

  function computeTarget(args::Dict{Symbol,Any},vmecSurface::VmecSurface{T},comm::MPI.Comm) where {T}
    if MPI.Comm_rank(comm) == 0
      return target_quasisymmetry(args[:mQS],args[:nQS],args[:mBoozer],args[:nBoozer],vmecSurface)
    else
      return nothing
    end
  end

end
