import PlasmaEquilibriumToolkit.gradB, PlasmaEquilibriumToolkit.B_norm, PlasmaEquilibriumToolkit.B_field

"""
    thetaVmec(x::FluxCoordinates,Λmn::VmecFourierData,interval=0.25)

Computes the internval VMEC θ coordinate associated with the points `x` specified
in `FluxCoordinates`, `Λmn` must be the `VmecFourierData` from the surface
specified by `x.ψ`.
"""
function thetaVmec(x::FluxCoordinates,Λmn::AbstractVector{VmecFourierData{T}},interval=0.25) where {T <: AbstractFloat}
  function residual(θ::T)
    return θ - x.θ + inverseTransform(FluxCoordinates(x.ψ,θ,x.ζ),Λmn)
  end

  bracket = (x.θ-interval,x.θ+interval)
  try
    return Roots.find_zero(residual,bracket,Roots.Order2())
  catch err
    if is(err,Roots.ConvergenceFailed) && attempt <= 5
      return thetaVmec(x,Λmn,2*interval)
    end
  end
end

#function thetaVmec(x::FluxCoordinates,Λmn::VmecFourierData)
#  Λ = inverseSineTransform(x,Λmn)
#  return thetaVmec(x,Λ)
#end

function thetaVmec(x::FluxCoordinates,
                   vmec::VmecSurface{T};
                  ) where {T <: AbstractFloat}
  return thetaVmec(x,vmec.λmn)
end

function thetaVmec(x::AbstractArray{FluxCoordinates},
                   vmec::VmecSurface{T};
                  ) where {T <: AbstractFloat}
  y = Array{Float64,ndims(x)}(undef,size(x))
  @batch for i in eachindex(x)
    y[i] = thetaVmec(x[i],vmec.λmn)
  end
  return y
end

function gradB(x::VmecCoordinates,
               ev::BasisVectors{T},
               vmec::VmecSurface{T};
              ) where {T}
  dBds = inverseTransform(x,vmec.bmn;deriv=:ds)
  dBdθ = inverseTransform(x,vmec.bmn;deriv=:dθ)
  dBdζ = inverseTransform(x,vmec.bmn;deriv=:dζ)
  return ev[:,1]*dBds .+ ev[:,2]*dBdθ .+ ev[:,3]*dBdζ
end

function B_norm(x::VmecCoordinates,
               vmec::VmecSurface{T};
              ) where {T}
  return inverseTransform(x,vmec.bmn)
end

function B_field(x::VmecCoordinates,
                vmec::VmecSurface{T};
               ) where {T}
  ∇x = basis_vectors(Contravariant(),x,vmec)
  return Bfield(Contravariant(),∇x)
end
